import {extract} from "./Utilities.js";

export enum Access {
    Failure,
    Success,
}

export interface Image_access {
    access: Access;
    image_URL: string;
}

// https://mariusschulz.com/blog/typing-functions-in-typescript
type Send = (ia: Image_access) => void;

export interface Open_Food_Facts_item {
    images: Array<Promise<Image_access>>;
    image_URLs: Array<string>;
    product_name: string;
}

// https://mariusschulz.com/blog/typing-functions-in-typescript
type Ready = (offi: Open_Food_Facts_item) => void;

export class Open_Food_Facts {

    private static readonly _All: Array<Open_Food_Facts> = new Array();

    static Get_instance(gtin: string): Open_Food_Facts {
        let off: Open_Food_Facts | undefined = Open_Food_Facts._All.find(element => {
            return element._gtin === gtin;
        });
        if (off === undefined) {
            // console.log("Created in 'Open_Food_Facts._All': " + gtin);
            off = new Open_Food_Facts(gtin);
            Open_Food_Facts._All.push(off);
        }
        return off;
    }

    private static readonly _URL: string = 'https://ssl-api.openfoodfacts.org';
    private static readonly _Path: string = '/api/v0/product/';
    private static readonly _RequestInit: RequestInit = {
        body: null, // See 'get' below
        headers: {
            'accept': 'application/json', // -> response
            // The only allowed values for the 'Content-Type' header are 'application/x-www-form-urlencoded', 'multipart/form-data' and 'text/plain'
            // https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
            // https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS#Simple_requests
            // "content-type": "application/json; charset=UTF-8" -> CORS problem
            'content-type': 'text/plain' // No need, anyway
        },
        method: 'get'
    };

    private static readonly _Status: string = 'status'; // '0' not found

    private readonly _item_data: Promise<Open_Food_Facts_item>;

    private constructor(private readonly _gtin: string) { // '_gtin' *field* automatic inclusion
        // => this._gtin = _gtin; -> no need!
        this._item_data = new Promise/*<Open_Food_Facts_item>*/((ready: Ready, problem) => { // 'problem' aims at being called when 'Promise' object has difficulty in achieving its job...
            this._get_Open_Food_Facts_item(ready, problem);
        });
    };

    // An 'async' function always returns a 'Promise' object (https://blog.bitsrc.io/keep-your-promises-in-typescript-using-async-await-7bdc57041308).
    // The 'Promise' object resolves by construction to the returned value:
    public async get_Open_Food_Facts_item() { // So returned type is inferred from 'async'...
        return await this._item_data; // Wait for '_get_item_data' (i.e., 'ready(offi);')
    }

    private _get_Open_Food_Facts_item(ready: Ready, problem, time_out?: number): void { // Call may omit 'time_out' optional parameter
        // https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API/Using_Fetch
        // https://en.wiki.openfoodfacts.org/API/Read/Product#Structure_of_the_call (e.g., https://ssl-api.openfoodfacts.org/api/v0/product/3664346300155.json):
        fetch(`${Open_Food_Facts._URL}${Open_Food_Facts._Path}${this._gtin}.json`, Open_Food_Facts._RequestInit)
            .then(function (response: Response) {
                response.json()
                    .then(function (result) {
                        console.assert(typeof result === 'object');
                        let found: number | undefined = result[Open_Food_Facts._Status];
                        console.log(this._gtin + ' status: ' + result[Open_Food_Facts._Status]);
                        if (found !== undefined && found === 1) {
                            let offi: Open_Food_Facts_item = {} as Open_Food_Facts_item;
                            offi['product_name'] = <unknown>extract(result, 'product', 'product_name') as string;
                            offi['image_URLs'] = [
                                <string><unknown>extract(result, 'product', 'image_url'),
                                <string><unknown>extract(result, 'product', 'image_front_url'),
                                <string><unknown>extract(result, 'product', 'image_small_url')
                            ];
                            // Remove 'null' and duplicates:
                            offi['image_URLs'] = Array.from(new Set(offi['image_URLs'].filter(image_URL => image_URL !== null)));
                            offi['images'] = [];
                            // Get images from URLs:
                            offi['image_URLs'].forEach(image_URL => {
                                console.log('Called URL: ' + image_URL);
                                // https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image
                                let candidate_image: HTMLImageElement = new Image;
                                // candidate_image.crossOrigin = "anonymous"; // Default
                                offi['images'].push(new Promise((send: Send) => {
                                    candidate_image.onload = () => {
                                        // console.assert(candidate_image.complete);
                                        send({access: Access.Success, image_URL: candidate_image.src});
                                    };
                                    candidate_image.onerror = () => {
                                        send({access: Access.Failure, image_URL: candidate_image.src});
                                    };
                                    candidate_image.src = image_URL;
                                }));

                                //     /**
                                //      * Fetch image is not permitted by CORS
                                //      */
                                //     let headers: Headers = new Headers();
                                //     headers.append('accept', 'image/*');
                                //     let options: RequestInit = {
                                //         headers: headers,
                                //         method: 'get'
                                //     };
                                //     fetch(image_URL, options)
                                //         .then(function (response: Response) {
                                //             response.blob()
                                //                 .then(function (blob: Blob) {
                                //                     // Process 'blob'
                                //                 }.bind(this))
                                //                 .catch(function (error: Error) {
                                //                     // This is where you run code if the server returns any errors
                                //                 });
                                //         }.bind(this))
                                //         .catch((error: Error) => {
                                //             window.alert('CORS error ' + image_URL + ': ' + error.message);
                                //         });
                                //     /**
                                //      * End of fetch image
                                //      */
                                //

                            }); // End of 'forEach(image_URL => {'
                            ready(offi);
                        } else { // Not found
                            let offi: Open_Food_Facts_item = {} as Open_Food_Facts_item;
                            offi['product_name'] = this._gtin;
                            offi['image_URLs'] = [];
                            offi['images'] = [];
                            // ready(offi);
                        }
                    }.bind(this))
                    .catch(function (error: Error) {
                        problem();
                        window.alert('Error response.json().then: ' + error.message);
                    });
            }.bind(this))
            .catch(function (error: Error) {
                problem();
                window.alert('Error ' + `${Open_Food_Facts._URL}: ` + error.message);
            });
    }
}

